#!/bin/bash

SOURCE_DIR=${PWD}
PACKAGE_DIR=/tmp/sicourse_package
PACKAGE_NAME=argos2-h414-extras
PACKAGE_VERSION=1cp
LOG_FILE=${SOURCE_DIR}/${PACKAGE_NAME}.log

function Error() {
    echo "A fatal error occurred. Check the log file ${LOG_FILE}."
    exit 1
}

function Usage() {
    echo "Usage: $0 [slack|debian|plain]"
    exit 1
}

function InstallDashel() {
    echo -n "Downloading dashel... "
    rm -rf source/dashel > /dev/null 2>> ${LOG_FILE} || Error
    svn co -r135 http://svn.gna.org/svn/dashel/trunk source/dashel > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
    echo -n "Patching dashel... "
    cd source/dashel > /dev/null 2>> ${LOG_FILE} || Error
    patch -p0 < ../dashel_r135.patch > /dev/null 2>> ${LOG_FILE} || Error
    cd - > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
    echo -n "Building dashel... "
    mkdir -p build/dashel > /dev/null 2>> ${LOG_FILE} || Error
    cd build/dashel > /dev/null 2>> ${LOG_FILE} || Error
    cmake -DCMAKE_INSTALL_PREFIX=/usr \
          -DCMAKE_BUILD_TYPE=Release \
          ../../source/dashel > /dev/null 2>> ${LOG_FILE} || Error
    make > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
    echo -n "Installing dashel... "
    sudo make DESTDIR=${PACKAGE_DIR}/debian install > /dev/null 2>> ${LOG_FILE} || Error
    cd ../.. > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
}

function InstallAseba() {
    echo -n "Downloading aseba... "
    rm -rf source/aseba > /dev/null 2>> ${LOG_FILE} || Error
    svn co -r438 http://svn.gna.org/svn/aseba/trunk source/aseba > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
    echo -n "Patching aseba... "
    cd source/aseba > /dev/null 2>> ${LOG_FILE} || Error
    patch -p1 < ../aseba_r442.patch > /dev/null 2>> ${LOG_FILE} || Error
    cd - > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
    echo -n "Building aseba... "
    mkdir -p build/aseba > /dev/null 2>> ${LOG_FILE} || Error
    cd build/aseba > /dev/null 2>> ${LOG_FILE} || Error
    cmake -DCMAKE_INSTALL_PREFIX=/usr \
        -DCMAKE_BUILD_TYPE=Release \
	      -DDASHEL_INCLUDE_DIR=${PACKAGE_DIR}/debian/usr/include \
	      -DDASHEL_LIBRARY=${PACKAGE_DIR}/debian/usr/lib/libdashel.a \
        ../../source/aseba > /dev/null 2>> ${LOG_FILE} || Error
    make > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
    echo -n "Installing aseba... "
    sudo make DESTDIR=${PACKAGE_DIR}/debian install > /dev/null 2>> ${LOG_FILE} || Error
    cd ../.. > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
}

function InstallEPuckARGoSCode() {
    echo -n "Building ARGoS+EPuck code... "
    mkdir -p build/argos2 > /dev/null 2>> ${LOG_FILE} || Error
    cd build/argos2 > /dev/null 2>> ${LOG_FILE} || Error
    cmake -DCMAKE_INSTALL_PREFIX=/usr \
          -DCMAKE_BUILD_TYPE=Release \
      	  -DDASHEL_LINK_DIR=${PACKAGE_DIR}/debian/usr/lib \
          -DASEBA_LINK_DIR=${PACKAGE_DIR}/debian/usr/lib \
          ../../source/argos2/code > /dev/null 2>> ${LOG_FILE} || Error
    make > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
    echo -n "Installing ARGoS+EPuck code... "
    sudo make DESTDIR=${PACKAGE_DIR}/debian install > /dev/null 2>> ${LOG_FILE} || Error
    cd ../.. > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
}

function AddResources() {
    echo -n "Adding resources... "
    sudo mkdir -p ${PACKAGE_DIR}/debian/usr/share/argos2/extra/H414_SwarmIntelligence > /dev/null 2>> ${LOG_FILE} || Error
    sudo cp -r source/argos2/resources/* ${PACKAGE_DIR}/debian/usr/share/argos2/extra/H414_SwarmIntelligence > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
}

function SlackPackage() {
    cd ${PACKAGE_DIR}/debian > /dev/null 2>> ${LOG_FILE} || Error
    echo -n "Stripping symbols... "
    find usr/bin -type f -print0 | xargs -0 sudo strip -s
    find usr/lib -type f -name '*.so' -print0 | xargs -0 sudo strip -s
    echo "OK"
    echo -n "Creating package... "
    sudo mkdir -p install
    sudo sh -c "cat <<'EOF' > install/slack-desc
                  |-----handy-ruler------------------------------------------------------|
argos2-h414-extras: ARGoS Extras for INFO-H-414 at ULB
argos2-h414-extras: 
argos2-h414-extras: This package contains extra material for use with the Swarm
argos2-h414-extras: Intelligence course INFO-H-414 at Universite Libre de Bruxelles.
argos2-h414-extras: 
argos2-h414-extras: 
argos2-h414-extras: 
argos2-h414-extras: 
argos2-h414-extras: 
argos2-h414-extras: 
argos2-h414-extras: Packager: Carlo Pinciroli <cpinciro@ulb.ac.be>
EOF
"
    sudo /sbin/makepkg -l y -c n ../${PACKAGE_NAME}-$(date +%Y%m%d)-$(uname -m)-${PACKAGE_VERSION}.txz > /dev/null 2>> ${LOG_FILE} || Error
    sudo mv ../${PACKAGE_NAME}-*.txz ${SOURCE_DIR} > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
}

function DebianPackage() {
    cd ${PACKAGE_DIR} > /dev/null 2>> ${LOG_FILE} || Error
    echo -n "Stripping symbols... "
    find debian/usr/bin -type f -print0 | xargs -0 sudo strip -s
    find debian/usr/lib -type f -name '*.so' -print0 | xargs -0 sudo strip -s
    echo "OK"
    echo -n "Creating package... "
    local PACKAGE_ARCH=$(dpkg --print-architecture)
    sudo mkdir -p debian debian/DEBIAN
    sudo sh -c "cat <<'EOF' > debian/DEBIAN/control
Package: ${PACKAGE_NAME}
Version: $(date +%Y%m%d)-${PACKAGE_VERSION}
Section: contrib/science
Priority: optional
Architecture: ${PACKAGE_ARCH}
Depends: argos2, libaudio2, libc6, libdbus-1-3, libexpat1, libfontconfig1, libfreetype6, libgcc1, libglib2.0-0, libhal1, libice6, libpcre3, libpng12-0, libqt4-dbus, libqt4-network, libqt4-svg, libqt4-xml, libqtcore4, libqtgui4, libqwt5-qt4, libsm6, libstdc++6, libuuid1, libx11-6, libxau6, libxcb1, libxdmcp6, libxext6, libxrender1, libxt6, zlib1g
Maintainer: Carlo Pinciroli <cpinciro@ulb.ac.be>
Description: ARGoS Extras for INFO-H-414 Swarm Intelligence Course
 This package contains extra material for use with the Swarm Intelligence
 course INFO-H-414 at Universite Libre de Bruxelles.
EOF
"
    find debian -name '*.so' -print0 | xargs -0 sudo chmod 644
    sudo dpkg-deb --build debian > /dev/null 2>> ${LOG_FILE} || Error
    sudo mv debian.deb ${SOURCE_DIR}/${PACKAGE_NAME}-$(date +%Y%m%d)-${PACKAGE_ARCH}-${PACKAGE_VERSION}.deb > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
}

function PlainPackage() {
    cd ${PACKAGE_DIR}/debian > /dev/null 2>> ${LOG_FILE} || Error
    echo -n "Stripping symbols... "
    find usr/bin -type f -print0 | xargs -0 sudo strip -s
    find usr/lib -type f -name '*.so' -print0 | xargs -0 sudo strip -s
    echo "OK"
    echo -n "Creating package... "
    sudo tar cf ${SOURCE_DIR}/${PACKAGE_NAME}-$(date +%Y%m%d)-$(uname -m)-${PACKAGE_VERSION}.tar.bz2 . > /dev/null 2>> ${LOG_FILE} || Error
    echo "OK"
}

# check mode
if [ $# -eq 1 ]; then
    MODE=$1
    if [ "x${MODE}" != "xslack" -a "x${MODE}" != "xdebian"  -a "x${MODE}" != "xplain" ]; then
        Usage
    fi
else
    Usage
fi

echo -n "(Re)creating necessary directories... "
sudo rm -rf ${PACKAGE_DIR} > /dev/null 2>> ${LOG_FILE} || Error
sudo mkdir -p ${PACKAGE_DIR}/debian > /dev/null 2>> ${LOG_FILE} || Error
rm -rf build > /dev/null 2>> ${LOG_FILE} || Error
mkdir -p build > /dev/null 2>> ${LOG_FILE} || Error
rm -f *.log
echo "OK"

InstallDashel
InstallAseba
InstallEPuckARGoSCode
AddResources
if [ "x${MODE}" = "xslack" ]; then
    SlackPackage
else if [ "x${MODE}" = "xdebian" ]; then
    DebianPackage
else
    PlainPackage
fi
fi

# cleanup
echo -n "Cleaning up... "
sudo rm -rf ${PACKAGE_DIR}
echo "OK"
