#include "epuck_aseba_controller.h"

#include <argos2/common/utility/math/general.h>
#include <algorithm>
#include <vector>

using namespace argos;

/****************************************/
/****************************************/

static CRange<Real> ENCODER_NOISE_RANGE(-0.3f, 0.3f);
typedef std::pair<UInt16, SEPuckRangeAndBearingReceivedPacket> TEPuckRABPacketPair;

/****************************************/
/****************************************/

bool ComparePackets(const TEPuckRABPacketPair& t_p1,
                    const TEPuckRABPacketPair& t_p2) {
   return (t_p1.second.Range < t_p2.second.Range);
}

/****************************************/
/****************************************/

CEPuckAsebaController::CEPuckAsebaController() :
   m_cRandomRange(0,10000),
   m_nState(0),
   m_nFoodItemIdx(-1) {
}

/****************************************/
/****************************************/

void CEPuckAsebaController::Init ( TConfigurationNode& t_tree ) {
   m_pcWheelsActuator= dynamic_cast<CCI_EPuckWheelsActuator*>(GetRobot().GetActuator("epuck_wheels"));
   m_pcProximitySensor= dynamic_cast<CCI_EPuckProximitySensor*>(GetRobot().GetSensor("epuck_proximity"));
   m_pcLedActuator = dynamic_cast<CCI_EPuckLedsActuator*>(GetRobot().GetActuator("epuck_leds"));
   m_pcGroundSensor = dynamic_cast<CCI_EPuckGroundSensor*>(GetRobot().GetSensor("epuck_ground"));
   m_pcLightSensor = dynamic_cast<CCI_EPuckLightSensor*>(GetRobot().GetSensor("epuck_light"));
   m_pcEncoderSensor = dynamic_cast<CCI_EPuckEncoderSensor*>(GetRobot().GetSensor("epuck_encoder"));
   m_pcRangeAndBearingActuator = dynamic_cast<CCI_EPuckRangeAndBearingActuator*>(GetRobot(  ).GetActuator( "epuck_range_and_bearing" ));
   m_pcRangeAndBearingSensor = dynamic_cast<CCI_EPuckRangeAndBearingSensor*>(GetRobot(  ).GetSensor( "epuck_range_and_bearing" ));
   m_pcRNG = CARGoSRandom::CreateRNG("argos");
   m_sVariables.State = 0;
   m_sVariables.Random = 0;
}

/****************************************/
/****************************************/

void CEPuckAsebaController::ControlStep (  ) {
   /* proximity sensor */
   for(UInt32 i = 0; i < CCI_EPuckProximitySensor::NUM_READINGS; ++i) {
      m_sVariables.Proximity[i] = static_cast<sint16>(1024.0f * m_pcProximitySensor->GetReading(i).Value);
   }

   /* light sensor */
   for(UInt32 i = 0; i < CCI_EPuckLightSensor::NUM_READINGS; ++i) {
      m_sVariables.Light[i] = static_cast<sint16>(1024.0f * m_pcLightSensor->GetReading(i).Value);
   }

   /* ground sensor */
   m_sVariables.Ground[0] = (m_pcGroundSensor->GetLeftReading() * 255);
   m_sVariables.Ground[1] = (m_pcGroundSensor->GetCentralReading() * 255);
   m_sVariables.Ground[2] = (m_pcGroundSensor->GetRightReading() * 255);

   /* encoder sensor
    * the values are relative to the step
    * so, a speed of 5 cm/s would be returned as 0.5cm per step (10 steps per sec)
    * thus, aseba would return 0
    * to solve this, we multiply by 10, in fact returning values in mm
    */
   CCI_EPuckEncoderSensor::SReading sEncoderReading = m_pcEncoderSensor->GetReading();
   m_sVariables.Encoder[0] = sEncoderReading.CoveredDistanceLeftWheel * 10.0f;
   m_sVariables.Encoder[1] = sEncoderReading.CoveredDistanceRightWheel * 10.0f;

   /* add noise */
   //m_sVariables.Encoder[0] += m_sVariables.Encoder[0] * m_pcRNG->Uniform(ENCODER_NOISE_RANGE);
   //m_sVariables.Encoder[1] += m_sVariables.Encoder[1] * m_pcRNG->Uniform(ENCODER_NOISE_RANGE);

   /* range & bearing sensor,
    * calculate the nearest 8 robots and put them in the variables
    */
   /* ranges are the distances wrt myself and the others, in centimeters */
   /* bearings are in degrees */
   CCI_EPuckRangeAndBearingSensor::TLastReceivedPackets tLastPackets = m_pcRangeAndBearingSensor->GetLastReceivedPackets();
   std::vector<TEPuckRABPacketPair> tDupPackets(tLastPackets.begin(), tLastPackets.end());
   UInt32 unPacketNum = Min<UInt32>(tDupPackets.size(), 8);
   std::partial_sort(tDupPackets.begin(),
                     tDupPackets.begin() + unPacketNum,
                     tDupPackets.end(),
                     ComparePackets);

   /* Zero all Aseba values */
   ::memset(m_sVariables.RABReceivedMsg, 0, sizeof(sint16) * 8);
   ::memset(m_sVariables.Range, 0, sizeof(sint16) * 8);
   ::memset(m_sVariables.Bearing, 0, sizeof(sint16) * 8);
      
   for (UInt32 i = 0; i < unPacketNum; ++i) {
      m_sVariables.Range[i] = static_cast<sint16>(tDupPackets.at(i).second.Range);
      m_sVariables.Bearing[i] = static_cast<sint16>(ToDegrees(tDupPackets.at(i).second.Bearing).GetValue());
      // convert two bytes to one int
      m_sVariables.RABReceivedMsg[i] = tDupPackets.at(i).second.Data[1];
      m_sVariables.RABReceivedMsg[i] *= 256;
      m_sVariables.RABReceivedMsg[i] += tDupPackets.at(i).second.Data[0];
   }
   m_pcRangeAndBearingSensor->ClearRABReceivedPackets();

   /* random number generator */
   m_sVariables.Random = m_pcRNG->Uniform(m_cRandomRange);

   /* prey */
   m_sVariables.Prey = (m_nFoodItemIdx == -1) ? 0 : 1;

   /* run VM */
   AsebaVMRun(&m_tVM, 65535);

   // reschedule a periodic event if we are not in step by step
   if (AsebaMaskIsClear(m_tVM.flags, ASEBA_VM_STEP_BY_STEP_MASK) ||
       AsebaMaskIsClear(m_tVM.flags, ASEBA_VM_EVENT_ACTIVE_MASK))
      AsebaVMSetupEvent(&m_tVM, ASEBA_EVENT_LOCAL_EVENTS_START);

   /* Wheels */
   m_pcWheelsActuator->SetLinearVelocity(m_sVariables.LSpeed, m_sVariables.RSpeed);

   /* LEDs */
   for (UInt32 i = 0; i < CCI_EPuckLedsActuator::NUM_LEDS; ++i)
      m_pcLedActuator->SwitchSingleLed( i, m_sVariables.Leds[i]);

   /* RAB */
   m_pcRangeAndBearingActuator->SetData(
      reinterpret_cast<const SEPuckRangeAndBearingReceivedPacket::TRangeAndBearingData&>
      (m_sVariables.RABSentMsg));

   /* State */
   m_nState = m_sVariables.State;
}

/****************************************/
/****************************************/

void CEPuckAsebaController::Reset() {
   m_nFoodItemIdx = -1;
}

/****************************************/
/****************************************/

REGISTER_CONTROLLER( CEPuckAsebaController, "epuck_aseba_controller" );
