#include "epuck_aseba_loop_functions.h"
#include "epuck_aseba_controller.h"
#include <argos2/common/utility/string_utilities.h>
#include <argos2/simulator/space/entities/floor_entity.h>
#include <argos2/simulator/space/entities/box_entity.h>

typedef std::map<AsebaVMState*, CEPuckAsebaController*>  VMEPuckMap;
static CEPuckAsebaLoopFunction* EPuckAsebaLoopFunction = NULL;

/* Aseba native function, used into aseba studio */
static AsebaNativeFunctionPointer NativeFunctions[] = {
   AsebaNative_veccopy,
   AsebaNative_vecfill,
   AsebaNative_vecadd,
   AsebaNative_vecsub,
   AsebaNative_vecmul,
   AsebaNative_vecdiv,
   AsebaNative_vecmin,
   AsebaNative_vecmax,
   AsebaNative_vecdot,
   AsebaNative_vecstat,
   AsebaNative_mathmuldiv,
   AsebaNative_mathatan2,
   AsebaNative_mathsin,
   AsebaNative_mathcos,
   AsebaNative_mathrot2,
   AsebaNative_mathsqrt,
   AsebaNative_lennardjones,
   AsebaNative_fermi
};

static const AsebaNativeFunctionDescription* NativeFunctionDescriptions[] = {
   &AsebaNativeDescription_veccopy,
   &AsebaNativeDescription_vecfill,
   &AsebaNativeDescription_vecadd,
   &AsebaNativeDescription_vecsub,
   &AsebaNativeDescription_vecmul,
   &AsebaNativeDescription_vecdiv,
   &AsebaNativeDescription_vecmin,
   &AsebaNativeDescription_vecmax,
   &AsebaNativeDescription_vecdot,
   &AsebaNativeDescription_vecstat,
   &AsebaNativeDescription_mathmuldiv,
   &AsebaNativeDescription_mathatan2,
   &AsebaNativeDescription_mathsin,
   &AsebaNativeDescription_mathcos,
   &AsebaNativeDescription_mathrot2,
   &AsebaNativeDescription_mathsqrt,
   &AsebaNativeDescription_lennardjones,
   &AsebaNativeDescription_fermi,
   0
};

extern AsebaVMDescription VMDescription;

/****************************************/
/****************************************/

/* Implementation of aseba glue code */
extern "C" void AsebaSendBuffer(AsebaVMState* pt_vm,
                                const uint8* pt_data,
                                uint16 un_length) {
   Dashel::Stream* ptStream = EPuckAsebaLoopFunction->m_ptStream;
   if (ptStream) {
      try {
         /* this may happen if target has disconnected */
         uint16 unLen = un_length - 2;
         ptStream->write(&unLen, 2);
         ptStream->write(&pt_vm->nodeId, 2);
         ptStream->write(pt_data, un_length);
         ptStream->flush();
      }
      catch (Dashel::DashelException e) {
         LOGERR << "[ASEBA] Cannot write to socket: " << e.what() << std::endl;
      }
   }

   EPuckAsebaLoopFunction->m_unLastMsgSource = pt_vm->nodeId;
   EPuckAsebaLoopFunction->m_vecLastMsgData.resize(un_length);
   ::memcpy(&EPuckAsebaLoopFunction->m_vecLastMsgData[0], pt_data, un_length);

   /* Make all the epucks process the incoming events */
   CSpace::TAnyEntityMap tMapEPucks = EPuckAsebaLoopFunction->GetSpace().GetEntitiesByType("epuck_entity");
   for (CSpace::TAnyEntityMap::iterator itEPuckEntity = tMapEPucks.begin();
        itEPuckEntity != tMapEPucks.end();
        ++itEPuckEntity) {
      CEPuckEntity& cEPuckEntity = *any_cast<CEPuckEntity*>(itEPuckEntity->second);
      CEPuckAsebaController& cEPuckController = dynamic_cast<CEPuckAsebaController&>(cEPuckEntity.GetControllableEntity().GetController());
      if (&cEPuckController.m_tVM != pt_vm) {
         AsebaProcessIncomingEvents(&cEPuckController.m_tVM);
      }
   }
}

extern "C" uint16 AsebaGetBuffer(AsebaVMState* pt_vm,
                                 uint8* pt_data,
                                 uint16 un_max_length,
                                 uint16* pt_source) {
   if (EPuckAsebaLoopFunction->m_vecLastMsgData.size()) {
      *pt_source = EPuckAsebaLoopFunction->m_unLastMsgSource;
      ::memcpy(pt_data, &EPuckAsebaLoopFunction->m_vecLastMsgData[0], EPuckAsebaLoopFunction->m_vecLastMsgData.size());
   }
   return EPuckAsebaLoopFunction->m_vecLastMsgData.size();
}

// this is really ugly, but a necessary hack
static char EPuckName[10] = "e-puck   ";

/* This is used by aseba to differentiate between the various e-puck, every e-puck has a specific name */
extern "C" const AsebaVMDescription* AsebaGetVMDescription(AsebaVMState* pt_vm)
{
   EPuckName[7] = '0' + pt_vm->nodeId / 10;
   EPuckName[8] = '0' + pt_vm->nodeId % 10;
   VMDescription.name = EPuckName;
   return &VMDescription;
}

static const AsebaLocalEventDescription LocalEvents[] = {
   { "timer", "periodic timer at 50 Hz" },
   { "ir_sensors", "New IR sensors values available" },
   { "camera", "New camera picture available" },
   { NULL, NULL }
};

extern "C" const AsebaLocalEventDescription * AsebaGetLocalEventsDescriptions(AsebaVMState* pt_vm) { return LocalEvents; }
extern "C" const AsebaNativeFunctionDescription * const * AsebaGetNativeFunctionsDescriptions(AsebaVMState* pt_vm) { return NativeFunctionDescriptions; }
extern "C" void AsebaNativeFunction(AsebaVMState* pt_vm, uint16 id) { NativeFunctions[id](pt_vm); }
extern "C" void AsebaWriteBytecode(AsebaVMState* pt_vm) {}
extern "C" void AsebaResetIntoBootloader(AsebaVMState* pt_vm) {}

extern "C" void AsebaAssert(AsebaVMState* pt_vm, AsebaAssertReason e_reason)
{
   LOGERR << std::endl << "[ASEBA] Fatal error: ";
   switch (pt_vm->nodeId)
   {
      case 1:  LOGERR << "left motor module"; break;
      case 2:  LOGERR << "right motor module"; break;
      case 3:  LOGERR << "proximity sensors module"; break;
      case 4:  LOGERR << "distance sensors module"; break;
      default: LOGERR << "unknown module"; break;
   }
   LOGERR << " has produced exception: ";
   switch (e_reason)
   {
      case ASEBA_ASSERT_UNKNOWN: LOGERR << "undefined"; break;
      case ASEBA_ASSERT_UNKNOWN_UNARY_OPERATOR:            LOGERR << "unknown unary operator"; break;
      case ASEBA_ASSERT_UNKNOWN_BINARY_OPERATOR:           LOGERR << "unknown binary operator"; break;
      case ASEBA_ASSERT_UNKNOWN_BYTECODE:                  LOGERR << "unknown bytecode"; break;
      case ASEBA_ASSERT_STACK_OVERFLOW:                    LOGERR << "stack overflow"; break;
      case ASEBA_ASSERT_STACK_UNDERFLOW:                   LOGERR << "stack underflow"; break;
      case ASEBA_ASSERT_OUT_OF_VARIABLES_BOUNDS:           LOGERR << "out of variables bounds"; break;
      case ASEBA_ASSERT_OUT_OF_BYTECODE_BOUNDS:            LOGERR << "out of bytecode bounds"; break;
      case ASEBA_ASSERT_STEP_OUT_OF_RUN:                   LOGERR << "step out of run"; break;
      case ASEBA_ASSERT_BREAKPOINT_OUT_OF_BYTECODE_BOUNDS: LOGERR << "breakpoint out of bytecode bounds"; break;
      default:                                             LOGERR << "unknown exception"; break;
   }
   LOGERR << "." << std::endl << "[ASEBA] pc = " << pt_vm->pc << ", sp = " << pt_vm->sp;
   LOGERR << std::endl << "[ASEBA] Resetting VM" << std::endl;
   assert(false);
   AsebaVMInit(pt_vm);
}

extern "C" void AsebaPutVmToSleep(AsebaVMState* pt_vm) {}

/* End of the aseba external c code */

/****************************************/
/****************************************/

CEPuckAsebaLoopFunction::CEPuckAsebaLoopFunction() :
   m_ptStream(NULL),
   m_bForaging(false),
   m_pcFloor(NULL),
   m_pcRNG(NULL),
   m_cForagingArenaSideX(-0.9f, 1.7f),
   m_cForagingArenaSideY(-1.7f, 1.7f),
   m_unCollectedFood(0),
   m_nEnergy(0),
   m_unEnergyPerFoodItem(1),
   m_unEnergyPerWalkingRobot(1),
   m_bConstruction(false) {
}

/****************************************/
/****************************************/

void CEPuckAsebaLoopFunction::Init(TConfigurationNode& t_tree) {
   try {
      ///// aseba studio part
      /* This initializes the single e-puck vm in the controller */
      /* Make all the epucks process the incoming events */
      m_tMapEPucks = m_cSpace.GetEntitiesByType("epuck_entity");
      for ( CSpace::TAnyEntityMap::iterator itEPuckEntity = m_tMapEPucks.begin();
            itEPuckEntity != m_tMapEPucks.end();
            ++itEPuckEntity) {
         CEPuckEntity& cEPuckEntity = *any_cast<CEPuckEntity*> (itEPuckEntity->second);
         CEPuckAsebaController& cEPuckController = dynamic_cast<CEPuckAsebaController&>(cEPuckEntity.GetControllableEntity().GetController());

         /* Getting the id from the last number found in the id string */
         std::string strId = (itEPuckEntity->first);
         std::string strDigits = "0123456789";
         UInt32 unEndPos = strId.find_last_of(strDigits);
         UInt32 unStartPos = unEndPos;
         while(strDigits.find(strId.at(unStartPos)) != std::string::npos) { // while the char is a number
            --unStartPos;
         }
         UInt32 unId = FromString<UInt32>(strId.substr(unStartPos + 1, unEndPos - unStartPos));

         /* Initializing the single e-puck vm in the controller */
         cEPuckController.m_tVM.nodeId = unId;
         cEPuckController.m_vecBytecode.resize(512);
         cEPuckController.m_tVM.bytecode = &cEPuckController.m_vecBytecode[0];
         cEPuckController.m_tVM.bytecodeSize = cEPuckController.m_vecBytecode.size();
         cEPuckController.m_vecStack.resize(64);
         cEPuckController.m_tVM.stack = &cEPuckController.m_vecStack[0];
         cEPuckController.m_tVM.stackSize = cEPuckController.m_vecStack.size();
         cEPuckController.m_tVM.variables = reinterpret_cast<sint16 *>(&cEPuckController.m_sVariables);
         cEPuckController.m_tVM.variablesSize = sizeof(cEPuckController.m_sVariables) / sizeof(sint16);
         AsebaVMInit(&cEPuckController.m_tVM);
         cEPuckController.m_sVariables.Id = unId;

         /* Making the variables avaiable for the external c code */
         EPuckAsebaLoopFunction = this;
      }

      /* Connecting to Aseba to get all the messages */
      LOG << "[INFO] Connection port: " << ASEBA_DEFAULT_PORT << std::endl;
      try {
         std::string strConnect = "tcpin:port=";
         strConnect += ToString(ASEBA_DEFAULT_PORT);
         Dashel::Hub::connect(strConnect);
      }
      catch (Dashel::DashelException& e) {
         THROW_ARGOSEXCEPTION_NESTED("Cannot create ASEBA listening port: " << ASEBA_DEFAULT_PORT, e);
      }
      if(NodeExists(t_tree, "foraging")) {
         m_bForaging = true;
         TConfigurationNode& tForaging = GetNode(t_tree, "foraging");
         /* Get a pointer to the floor entity */
         m_pcFloor = &m_cSpace.GetFloorEntity();
         /* Get the number of food items we want to be scattered from XML */
         UInt32 unFoodItems;
         GetNodeAttribute(tForaging, "items", unFoodItems);
         /* Get the number of food items we want to be scattered from XML */
         GetNodeAttribute(tForaging, "radius", m_fFoodSquareRadius);
         m_fFoodSquareRadius *= m_fFoodSquareRadius;
         /* Create a new RNG */
         m_pcRNG = CARGoSRandom::CreateRNG("argos");
         /* Distribute uniformly the items in the environment */
         for(UInt32 i = 0; i < unFoodItems; ++i) {
            m_cFoodPos.push_back(
               CVector2(m_pcRNG->Uniform(m_cForagingArenaSideX),
                        m_pcRNG->Uniform(m_cForagingArenaSideY)));
         }
         /* Get the output file name from XML */
         GetNodeAttribute(tForaging, "output", m_strOutput);
         /* Open the file, erasing its contents */
         m_cOutput.open(m_strOutput.c_str(), std::ios_base::trunc | std::ios_base::out);
         m_cOutput << "# clock\twalking\tresting\tcollected_food\tenergy" << std::endl;
         /* Get energy gain per item collected */
         GetNodeAttribute(tForaging, "energy_per_item", m_unEnergyPerFoodItem);
         /* Get energy loss per walking robot */
         GetNodeAttribute(tForaging, "energy_per_walking_robot", m_unEnergyPerWalkingRobot);
      }

      if(NodeExists(t_tree, "construction")) {
         m_bConstruction = true;
         TConfigurationNode& tConstruction = GetNode(t_tree, "construction");
         /* Get the output file name from XML */
         GetNodeAttribute(tConstruction, "output", m_strOutput);
         /* Open the file, erasing its contents */
         m_cOutput.open(m_strOutput.c_str(), std::ios_base::trunc | std::ios_base::out);
         m_cOutput << "# time\tid\tposx\tposy" << std::endl;
      }
   }
   catch(CARGoSException& ex) {
      THROW_ARGOSEXCEPTION_NESTED("Error initializing the experiment.", ex);
   }
}

/****************************************/
/****************************************/

void CEPuckAsebaLoopFunction::Reset() {
   if(m_bForaging) {
      /* Zero the counters */
      m_unCollectedFood = 0;
      m_nEnergy = 0;
      /* Close the file */
      m_cOutput.close();
      /* Open the file, erasing its contents */
      m_cOutput.open(m_strOutput.c_str(), std::ios_base::trunc | std::ios_base::out);
      m_cOutput << "# clock\twalking\tresting\tcollected_food\tenergy" << std::endl;
      /* Distribute uniformly the items in the environment */
      for(UInt32 i = 0; i < m_cFoodPos.size(); ++i) {
         m_cFoodPos[i].Set(m_pcRNG->Uniform(m_cForagingArenaSideX),
                           m_pcRNG->Uniform(m_cForagingArenaSideY));
      }
   }
   if(m_bConstruction) {
      /* Close the file */
      m_cOutput.close();
      /* Open the file, erasing its contents */
      m_cOutput.open(m_strOutput.c_str(), std::ios_base::trunc | std::ios_base::out);
      m_cOutput << "# time\tid\tposx\tposy" << std::endl;
   }
}

/****************************************/
/****************************************/

void CEPuckAsebaLoopFunction::Destroy() {
   if(m_bForaging || m_bConstruction) {
      /* Close the file */
      m_cOutput.close();
   }
}

/****************************************/
/****************************************/

void CEPuckAsebaLoopFunction::PrePhysicsEngineStep() {
   /* Do an aseba (dashel) network step to retrive the data */
   Hub::step();
   if(m_bForaging) {
      /* Logic to pick and drop food items */
      /*
       * If a robot is in the nest, drop the food item
       * If a robot is on a food item, pick it
       * Each robot can carry only one food item per time
       */
      UInt32 unWalkingEPucks = 0;
      UInt32 unRestingEPucks = 0;
      CSpace::TAnyEntityMap& m_cEPucks = m_cSpace.GetEntitiesByType("epuck_entity");
      for(CSpace::TAnyEntityMap::iterator it = m_cEPucks.begin();
          it != m_cEPucks.end();
          ++it) {
         /* Get handle to e-puck entity and controller */
         CEPuckEntity& cEPuckEntity = *any_cast<CEPuckEntity*> (it->second);
         CEPuckAsebaController& cEPuckController = dynamic_cast<CEPuckAsebaController&>(cEPuckEntity.GetControllableEntity().GetController());
         /* Count how many e-pucks are in which state */
         if(cEPuckController.m_nState == 0) ++unWalkingEPucks;
         else ++unRestingEPucks;
         /* Set the position of the e-puck on the ground */
         CVector2 cEPuckPos;
         cEPuckPos.Set(cEPuckEntity.GetEmbodiedEntity().GetPosition().GetX(),
                       cEPuckEntity.GetEmbodiedEntity().GetPosition().GetY());
         /* The e-puck has a food item */
         if(cEPuckController.m_nFoodItemIdx >= 0) {
            /* Check whether the e-puck is in the nest */
            if(cEPuckPos.GetX() < -1.0f) {
               /* Place a new food item in the ground */
               m_cFoodPos[cEPuckController.m_nFoodItemIdx].Set(m_pcRNG->Uniform(m_cForagingArenaSideX),
                                                               m_pcRNG->Uniform(m_cForagingArenaSideY));
               /* Drop the food item */
               cEPuckController.m_nFoodItemIdx = -1;
               /* Increase the energy and food count */
               m_nEnergy += m_unEnergyPerFoodItem;
               ++m_unCollectedFood;
               /* The floor texture must be updated */
               m_pcFloor->SetChanged();
            }
         }
         else {
            /* The e-puck has no food item */
            /* Check whether the e-puck is out of the nest */
            if(cEPuckPos.GetX() > -1.0f) {
               /* Check whether the e-puck is on a food item */
               bool bDone = false;
               for(UInt32 i = 0; i < m_cFoodPos.size() && !bDone; ++i) {
                  if((cEPuckPos - m_cFoodPos[i]).SquareLength() < m_fFoodSquareRadius) {
                     /* If so, we move that item out of sight */
                     m_cFoodPos[i].Set(100.0f, 100.f);
                     /* The e-puck is now carrying an item */
                     cEPuckController.m_nFoodItemIdx = i;
                     /* The floor texture must be updated */
                     m_pcFloor->SetChanged();
                     /* We are done */
                     bDone = true;
                  }
               }
            }
         }
      }
      /* Update energy expediture due to walking robots */
      m_nEnergy -= unWalkingEPucks * m_unEnergyPerWalkingRobot;
      /* Output stuff to file */
      m_cOutput << m_cSpace.GetSimulationClock() << "\t"
                << unWalkingEPucks << "\t"
                << unRestingEPucks << "\t"
                << m_unCollectedFood << "\t"
                << m_nEnergy << std::endl;
   }
   /* Construction-related logging */
   if(m_bConstruction && m_cSpace.GetSimulationClock() % 10 == 0) {
      CSpace::TAnyEntityMap& m_cObjects = m_cSpace.GetEntitiesByType("box_entity");
      for(CSpace::TAnyEntityMap::iterator it = m_cObjects.begin();
          it != m_cObjects.end();
          ++it) {
         CBoxEntity& cBox = *any_cast<CBoxEntity*> (it->second);
         if(cBox.GetId().compare(0, 2, "ob") == 0) {
            m_cOutput << m_cSpace.GetSimulationClock() << "\t"
                      << cBox.GetId() << "\t"
                      << cBox.GetEmbodiedEntity().GetPosition().GetX() << "\t"
                      << cBox.GetEmbodiedEntity().GetPosition().GetY() << "\t"
                      << std::endl;
         }
      }
   }
}

/****************************************/
/****************************************/

/* Connect the loop function to aseba studio, via dashel */

void CEPuckAsebaLoopFunction::connectionCreated(Dashel::Stream* pt_stream)
{
   std::string strTargetName = pt_stream->getTargetName();
   if (strTargetName.substr(0, strTargetName.find_first_of(':')) == "tcp")
   {
      LOGERR << "[ASEBA] New client connected." << std::endl;
      if (m_ptStream)
      {
         closeStream(m_ptStream);
         LOGERR << "[ASEBA] Disconnected old client." << std::endl;
      }
      m_ptStream = pt_stream;
   }
}

/****************************************/
/****************************************/
/* Get the incoming data from the simulator, and distribute it to the right e-puck, then make the e-puck process them */

void CEPuckAsebaLoopFunction::incomingData(Dashel::Stream* pt_stream) {
   uint16 unLen;
   uint16 unIncomingMsgSource;
   std::valarray<uint8> vecIncomingMsgData;

   pt_stream->read(&unLen, 2);
   pt_stream->read(&unIncomingMsgSource, 2);
   vecIncomingMsgData.resize(unLen + 2);
   pt_stream->read(&vecIncomingMsgData[0], vecIncomingMsgData.size());
   for(CSpace::TAnyEntityMap::iterator itEPuckEntity = m_tMapEPucks.begin();
       itEPuckEntity != m_tMapEPucks.end();
       ++itEPuckEntity) {
      CEPuckEntity& cEPuckEntity = *any_cast<CEPuckEntity*>(itEPuckEntity->second);
      CEPuckAsebaController& ePuckController = dynamic_cast<CEPuckAsebaController&>( cEPuckEntity.GetControllableEntity().GetController());
      m_unLastMsgSource = unIncomingMsgSource;
      m_vecLastMsgData.resize(vecIncomingMsgData.size());
      ::memcpy(&m_vecLastMsgData[0],
               &vecIncomingMsgData[0],
               vecIncomingMsgData.size());
      AsebaProcessIncomingEvents(&(ePuckController.m_tVM));
   }
}

/****************************************/
/****************************************/

/* Close the connections and reset all the vms */
void CEPuckAsebaLoopFunction::connectionClosed(Dashel::Stream* pt_stream,
                                               bool b_abnormal) {
   if (pt_stream == m_ptStream) {
      m_ptStream = 0;
			// clear breakpoints
      for (CSpace::TAnyEntityMap::iterator itEPuckEntity = m_tMapEPucks.begin();
           itEPuckEntity != m_tMapEPucks.end();
           ++itEPuckEntity) {
         CEPuckEntity& cEPuckEntity = *any_cast<CEPuckEntity*>(itEPuckEntity->second);
         CEPuckAsebaController& cEPuckController = dynamic_cast<CEPuckAsebaController&>(cEPuckEntity.GetControllableEntity().GetController());
         (cEPuckController.m_tVM).breakpointsCount = 0;
      }
   }
   if(b_abnormal) {
      LOGERR << "[ASEBA] Client has disconnected unexpectedly." << std::endl;
   }
   else {
      LOGERR << "[ASEBA] Client has disconnected properly." << std::endl;
   }
}

/****************************************/
/****************************************/

CColor CEPuckAsebaLoopFunction::GetFloorColor(const CVector2& c_position_on_plane) {
   if(m_bForaging) {
      if(c_position_on_plane.GetX() < -1.0f) {
         return CColor::GRAY50;
      }
      for(UInt32 i = 0; i < m_cFoodPos.size(); ++i) {
         if((c_position_on_plane - m_cFoodPos[i]).SquareLength() < m_fFoodSquareRadius) {
            return CColor::BLACK;
         }
      }
   }
   return CColor::WHITE;
}

/****************************************/
/****************************************/

REGISTER_LOOP_FUNCTIONS(CEPuckAsebaLoopFunction, "epuck_aseba_loop_functions");
